/*
 * Copyright 2016 - 2024 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#pragma once

#include <studio/studio.hpp>

#include <nostalgia/core/color.hpp>
#include <nostalgia/core/gfx.hpp>
#include <nostalgia/core/palette.hpp>

namespace nostalgia::core {

enum class PaletteEditorCommandId {
	AddPage,
	DuplicatePage,
	RemovePage,
	AddColor,
	RemoveColor,
	UpdateColor,
	MoveColor,
};


class AddPageCommand: public studio::UndoCommand {
	private:
		Palette &m_pal;

	public:
		AddPageCommand(Palette &pal) noexcept;

		~AddPageCommand() noexcept override = default;

		[[nodiscard]]
		int commandId() const noexcept final;

		void redo() noexcept final;

		void undo() noexcept final;

};

class DuplicatePageCommand: public studio::UndoCommand {
	private:
		Palette &m_pal;
		size_t m_dstIdx = 0;
		ox::Vector<Color16> m_page;

	public:
		DuplicatePageCommand(Palette &pal, size_t srcIdx, size_t dstIdx) noexcept;

		~DuplicatePageCommand() noexcept override = default;

		[[nodiscard]]
		int commandId() const noexcept final;

		void redo() noexcept final;

		void undo() noexcept final;

		[[nodiscard]]
		size_t insertIdx() const noexcept;

};

class RemovePageCommand: public studio::UndoCommand {
	private:
		Palette &m_pal;
		size_t m_idx = 0;
		ox::Vector<Color16> m_page;

	public:
		RemovePageCommand(Palette &pal, size_t idx) noexcept;

		~RemovePageCommand() noexcept override = default;

		[[nodiscard]]
		int commandId() const noexcept final;

		void redo() noexcept final;

		void undo() noexcept final;

};

class AddColorCommand: public studio::UndoCommand {
	private:
		Palette *m_pal = nullptr;
		Color16 m_color = 0;
		int m_idx = -1;
		size_t m_page = 0;

	public:
		AddColorCommand(Palette *pal, Color16 color, size_t page, int idx) noexcept;

		~AddColorCommand() noexcept override = default;

		[[nodiscard]]
		int commandId() const noexcept override;

		void redo() noexcept override;

		void undo() noexcept override;

};

class RemoveColorCommand: public studio::UndoCommand {
	private:
		Palette *m_pal = nullptr;
		Color16 m_color = 0;
		size_t m_page = 0;
		int m_idx = -1;

	public:
		RemoveColorCommand(Palette *pal, Color16 color, size_t page, int idx) noexcept;

		~RemoveColorCommand() noexcept override = default;

		[[nodiscard]]
		int commandId() const noexcept override;

		void redo() noexcept override;

		void undo() noexcept override;

};

class UpdateColorCommand: public studio::UndoCommand {
	private:
		Palette *m_pal = nullptr;
		Color16 m_oldColor = 0;
		Color16 m_newColor = 0;
		size_t m_page = 0;
		int m_idx = -1;

	public:
		UpdateColorCommand(Palette *pal, size_t page, int idx, Color16 oldColor, Color16 newColor) noexcept;

		~UpdateColorCommand() noexcept override = default;

		[[nodiscard]]
		bool mergeWith(const UndoCommand *cmd) noexcept final;

		[[nodiscard]]
		int commandId() const noexcept final;

		void redo() noexcept final;

		void undo() noexcept final;

};

class MoveColorCommand: public studio::UndoCommand {
	private:
		Palette *m_pal = nullptr;
		size_t m_page = 0;
		std::size_t m_idx = 0;
		int m_offset = 0;

	public:
		MoveColorCommand(Palette *pal, size_t page, std::size_t idx, int offset) noexcept;

		~MoveColorCommand() noexcept override = default;

		[[nodiscard]]
		int commandId() const noexcept override;

	public:
		void redo() noexcept override;

		void undo() noexcept override;

	private:
		void moveColor(int idx, int offset) noexcept;
};

}
